package de.howaner.MapEditor.gui.components;

import de.howaner.MapEditor.MainMapEditor;
import de.howaner.MapEditor.map.MapImageSerializer;
import de.howaner.MapEditor.map.MapTerrain;
import de.howaner.MapEditor.map.MapWalkData;
import de.howaner.MapEditor.map.events.Event;
import de.howaner.MapEditor.map.events.EventType;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Map;
import javax.swing.ImageIcon;
import javax.swing.JLabel;

public class GuiImageMap extends JLabel implements MouseListener {

	private int zoomLevel = 1;
	private Thread inZoomThread = null;
	private Map<BufferedImage, BufferedImage> scaledCache = new HashMap<BufferedImage, BufferedImage>();
	private BufferedImage scaledImage;
	private BufferedImage image;
	private Thread mouseThread;
	private boolean isLeftClick = false;
	private Event selectedEvent = null;

	public GuiImageMap() {
		this.addMouseListener(this);
		this.reload();
	}

	public void drawWalkData() {
		MapTerrain map = MainMapEditor.instance.getLoadedMap();

		for (int x = 0; x < map.getWidth(); x++) {
			for (int y = 0; y < map.getHeight(); y++) {
				MapWalkData.WalkDataType type = map.getWalkData().getWalkData(x, y);
				if (type != MapWalkData.WalkDataType.NONE)
					this.drawImage(type.getImage(), x, y, false);
			}
		}
	}

	public void drawEvents() {
		for (int x = 0; x < MainMapEditor.instance.getLoadedMap().getWidth(); x++) {
			for (int y = 0; y < MainMapEditor.instance.getLoadedMap().getHeight(); y++) {
				Event event = MainMapEditor.instance.getLoadedMap().getEvents().getEventAt(x, y);
				if (event != null) {
					if (this.selectedEvent == event) {
						this.drawImage(event.getType().getSelectedImage(), x, y, false);
					} else {
						this.drawImage(event.getType().getImage(), x, y, false);
					}
				}
			}
		}
	}

	/**
	 * This function will draw the image to the this.scaledImage image
	 * @param image The image to draw
	 * @param x The x position (tile positions !!)
	 * @param y The y position (tile positions !!)
	 * @param imageOverride Should the image override the complete area?
	 */
	public void drawImage(BufferedImage image, int x, int y, boolean imageOverride) {
		y = MainMapEditor.instance.getLoadedMap().getHeight() - 1 - y;

		BufferedImage imageToDraw = this.scaledCache.get(image);
		if (imageToDraw == null) {
			imageToDraw = zoom(image, this.zoomLevel);
			this.scaledCache.put(image, imageToDraw);
		}

		// Write to scaled image
		for (int tileX = 0; tileX < 16 * this.zoomLevel; tileX++) {
			for (int tileY = 0; tileY < 16 * this.zoomLevel; tileY++) {
				int RGB = imageToDraw.getRGB(tileX, tileY);
				int alpha = (RGB >> 24) & 0xff;
				if (alpha != 0) {
					this.scaledImage.setRGB(x * 16 * this.zoomLevel + tileX, y * 16 * this.zoomLevel + tileY, RGB);
				} else if (imageOverride) {
					this.scaledImage.setRGB(x * 16 * this.zoomLevel + tileX, y * 16 * this.zoomLevel + tileY, this.image.getRGB(x * 16 + (tileX / this.zoomLevel), y * 16 + (tileY / this.zoomLevel)));
				}
			}
		}
	}

	public void doZoom(int modifier) {
		if (this.inZoomThread != null) {
			return;
		}

		this.zoomLevel = Math.min(4, Math.max(this.zoomLevel + modifier, 1));
		this.scaledCache.clear();

		this.inZoomThread = new Thread(new Runnable() {
			@Override
			public void run() {
				BufferedImage zoomedImage = GuiImageMap.zoom(GuiImageMap.this.image, GuiImageMap.this.zoomLevel);
				ImageIcon icon = new ImageIcon(zoomedImage);
				synchronized (GuiImageMap.this) {
					GuiImageMap.this.scaledImage = zoomedImage;
					if (GuiImageMap.this.isWalkDataSelected()) {
						GuiImageMap.this.drawWalkData();
					} else if (GuiImageMap.this.isEventSelected()) {
						GuiImageMap.this.drawEvents();
					}
					GuiImageMap.this.setIcon(icon);
					GuiImageMap.this.inZoomThread = null;
				}
			}
		});
		this.inZoomThread.setName("Map Zoom thread");
		this.inZoomThread.start();
	}

	public void reload() {
		this.image = MapImageSerializer.serializeMapToImage(MainMapEditor.instance.getLoadedMap());
		this.doZoom(0);
	}

	public static BufferedImage zoom(BufferedImage image, int zoomLevel) {
		int newWidth = image.getWidth() * zoomLevel;
		int newHeight = image.getHeight() * zoomLevel;

		double scaleX = (double) (newWidth / image.getWidth());
		double scaleY = (double) (newHeight / image.getHeight());
		AffineTransform scaleTransform = AffineTransform.getScaleInstance(scaleX, scaleY);
		AffineTransformOp bilinearScaleOp = new AffineTransformOp(scaleTransform, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);

		return bilinearScaleOp.filter(image, new BufferedImage(newWidth, newHeight, image.getType()));
	}
	
	public void clicked(int x, int y) {
		if (this.isWalkDataSelected()) {
			MapWalkData.WalkDataType walkData = MainMapEditor.instance.getWindow().getSelectedWalkData();
			if (walkData != null) {
				MainMapEditor.instance.getLoadedMap().getWalkData().setWalkData(x, y, walkData);
				this.drawImage(walkData.getImage(), x, y, true);
				this.setIcon(new ImageIcon(this.scaledImage));
			}
		} else if (this.isEventSelected()) {
			Event event = MainMapEditor.instance.getLoadedMap().getEvents().getEventAt(x, y);
			if (event != null) {
				this.selectEvent(event);
				return;
			}

			EventType eventType = MainMapEditor.instance.getWindow().getSelectedEventType();
			if (eventType != null) {
				event = eventType.createEventInstance(MainMapEditor.instance.getLoadedMap(), x, y);
				MainMapEditor.instance.getLoadedMap().getEvents().setEvent(x, y, event);
				this.drawImage(eventType.getImage(), x, y, true);
				this.setIcon(new ImageIcon(this.scaledImage));
			}
		}
	}

	@Override
	public void mouseClicked(MouseEvent me) { }

	@Override
	public void mousePressed(MouseEvent me) {
		if (me.getButton() == 1) {
			// Left click
			this.isLeftClick = true;

			if (this.isWalkDataSelected() && MainMapEditor.instance.getWindow().getSelectedWalkData() != null) {
				this.startMouseThread();
			} else if (this.isEventSelected()) {
				this.clicked(me.getX() / this.zoomLevel / 16, MainMapEditor.instance.getLoadedMap().getHeight() - 1 - (me.getY() / this.zoomLevel / 16));
			}
		} else if (me.getButton() == 3) {
			// Right click
			if (this.isWalkDataSelected()) {
				// ...
			} else if (this.isEventSelected()) {
				int x = me.getX() / this.zoomLevel / 16;
				int y = MainMapEditor.instance.getLoadedMap().getHeight() - 1 - (me.getY() / this.zoomLevel / 16);

				Event event = MainMapEditor.instance.getLoadedMap().getEvents().getEventAt(x, y);
				if (event != null) {
					if (this.selectedEvent == event) {
						MainMapEditor.instance.getWindow().selectEvent(null);
						this.selectedEvent = null;
					}

					MainMapEditor.instance.getLoadedMap().getEvents().setEvent(x, y, null);
					this.drawImage(MapWalkData.WalkDataType.NONE.getImage(), x, y, true);
					this.setIcon(new ImageIcon(this.scaledImage));
				}
			}
		}
	}

	public void selectEvent(Event event) {
		MainMapEditor.instance.getWindow().selectEvent(event);

		if (this.selectedEvent != null) {
			this.drawImage(this.selectedEvent.getType().getImage(), this.selectedEvent.getPosition().getPosX(), this.selectedEvent.getPosition().getPosY(), true);
			this.selectedEvent = null;
		}

		if (event == null) {
			MainMapEditor.instance.getWindow().selectEvent(null);
			return;
		}

		this.selectedEvent = event;
		this.drawImage(event.getType().getSelectedImage(), event.getPosition().getPosX(), event.getPosition().getPosY(), false);
		this.setIcon(new ImageIcon(this.scaledImage));
	}

	public void startMouseThread() {
		if (this.mouseThread != null && this.mouseThread.isAlive()) {
			this.mouseThread.interrupt();
		}

		this.mouseThread = new Thread(new Runnable() {
			private int lastPositionX = -1;
			private int lastPositionY = -1;

			@Override
			public void run() {
				while (!Thread.currentThread().isInterrupted() && GuiImageMap.this.isLeftClick) {
					Point coords = GuiImageMap.this.getMousePosition();
					if (coords == null) {
						continue;
					}

					int x = coords.x / GuiImageMap.this.zoomLevel / 16;
					int y = MainMapEditor.instance.getLoadedMap().getHeight() - 1 - (coords.y / GuiImageMap.this.zoomLevel / 16);

					if (this.lastPositionX == x && this.lastPositionY == y) {
						continue;
					}

					this.lastPositionX = x;
					this.lastPositionY = y;
					synchronized (GuiImageMap.this) {
						GuiImageMap.this.clicked(x, y);
					}
				}
			}
		});
		this.mouseThread.setName("Mouse finder");
		this.mouseThread.start();
	}

	@Override
	public void mouseReleased(MouseEvent me) {
		if (me.getButton() == 1) {
			this.isLeftClick = false;

			if (this.mouseThread != null) {
				this.mouseThread.interrupt();
				this.mouseThread = null;
			}
		}
	}

	@Override
	public void mouseExited(MouseEvent me) {
		if (this.mouseThread != null) {
			this.mouseThread.interrupt();
			this.mouseThread = null;
		}
	}

	@Override
	public void mouseEntered(MouseEvent me) {
		if (!this.isWalkDataSelected()) {
			return;
		}

		if (me.getPoint() != null && this.isLeftClick && this.isWalkDataSelected() && MainMapEditor.instance.getWindow().getSelectedWalkData() != null) {
			this.startMouseThread();
		}
	}

	public Event getSelectedEvent() {
		return this.selectedEvent;
	}

	public Thread getCurrentZoomThread() {
		return this.inZoomThread;
	}

	public boolean isWalkDataSelected() {
		return (MainMapEditor.instance.getWindow() == null || MainMapEditor.instance.getWindow().getSelectedTabID() == 0);
	}

	public boolean isEventSelected() {
		return (MainMapEditor.instance.getWindow() != null && MainMapEditor.instance.getWindow().getSelectedTabID() == 1);
	}

}
