package de.howaner.MapEditor.map.tileset;

import de.howaner.MapEditor.util.Base64;
import de.howaner.Pokemon.saveformat.TagBase;
import de.howaner.Pokemon.saveformat.TagCompound;
import de.howaner.Pokemon.saveformat.TagString;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import javax.imageio.ImageIO;

public class MapTileset {
	private List<BufferedImage> tiles = new ArrayList<BufferedImage>();
	private final Map<Short, TileAnimation> animations = new HashMap<Short, TileAnimation>();
	private int tilesPerLine = 16;

	public MapTileset() { }

	public MapTileset(List<BufferedImage> tiles) {
		this.tiles = tiles;
	}

	public int getTilesPerLine() {
		return this.tilesPerLine;
	}

	public void setTilesPerLine(int tiles) {
		this.tilesPerLine = tiles;
	}

	public List<BufferedImage> getTiles() {
		return this.tiles;
	}

	public void setTiles(List<BufferedImage> tiles) {
		this.tiles = tiles;
	}

	public void addTile(BufferedImage tile) {
		this.tiles.add(tile);
	}

	public void removeTile(BufferedImage tile) {
		this.tiles.remove(tile);
	}

	public BufferedImage getTileFromId(int id) {
		if (id < 0 || id >= tiles.size()) {
			return null;
		}

		return this.tiles.get(id);
	}

	public void setTile(int id, BufferedImage tile) {
		this.tiles.set(id, tile);
	}

	public TileAnimation getTileAnimation(short tile) {
		return this.animations.get(tile);
	}

	public void setTileAnimation(short tile, TileAnimation animation) {
		if (tile >= this.tiles.size()) {
			return;
		}

		this.animations.put(tile, animation);
	}

	public void loadFromCompound(TagCompound compound) {
		this.tiles.clear();
		List<TagBase> tilesList = compound.getList("tiles");
		for (TagBase tag : tilesList) {
			if (!(tag instanceof TagString)) {
				continue;
			}

			String base64String = ((TagString)tag).getData();
			try {
				this.tiles.add(decodeBase64(base64String));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		if (compound.containsTag("TilesPerLine"))
			this.tilesPerLine = compound.getInt("TilesPerLine");

		if (compound.containsTag("animations")) {
			TagCompound animationsCompound = compound.getCompound("animations");
			for (String key : animationsCompound.getKeys()) {
				short id = Short.parseShort(key);
				TileAnimation animation = new TileAnimation();

				try {
					animation.loadFromCompound(animationsCompound.getCompound(key));
				} catch (Exception e) {
					System.out.println("Can't load tileset animation #" + id);
					e.printStackTrace();
					continue;
				}

				this.setTileAnimation(id, animation);
			}
		}
	}

	public void saveToCompound(TagCompound compound) {
		List<TagBase> tilesList = new ArrayList<TagBase>();

		for (BufferedImage tile : this.tiles) {
			try {
				tilesList.add(new TagString(encodeBase64(tile)));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		compound.setList("tiles", 7, tilesList);

		compound.setInt("TilesPerLine", this.tilesPerLine);

		if (!this.animations.isEmpty()) {
			TagCompound animationsCompound = new TagCompound();
			for (Entry<Short, TileAnimation> e : this.animations.entrySet()) {
				TagCompound animationCompound = new TagCompound();
				try {
					e.getValue().saveToCompound(animationCompound);
				} catch (Exception ex) {
					System.out.println("Can't save tileset animation #" + e.getKey());
					ex.printStackTrace();
					continue;
				}

				animationsCompound.setCompound(e.getKey().toString(), animationCompound);
			}

			compound.setCompound("animations", animationsCompound);
		}
	}

	public static BufferedImage decodeBase64(String s) throws Exception {
		byte[] bytes = Base64.decode(s, Base64.DEFAULT);
		return ImageIO.read(new ByteArrayInputStream(bytes));
	}

	public static String encodeBase64(BufferedImage image) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ImageIO.write(image, "png", baos);
		baos.flush();
		byte[] bytes = baos.toByteArray();
		baos.close();

		return Base64.encodeToString(bytes, Base64.DEFAULT);
	}

}
